import numpy as np
import os
#import os.path as osp
import yaml
import matplotlib.pyplot as plt
from scipy.interpolate import PchipInterpolator
import wisdem.inputs as sch    # used for loading turbine YAML and using WISDEM validation process
from wisdem.commonse.utilities import arc_length

# test local code; consider src layout in future to test installed code
import raft as raft
import moorpy as mp


import importlib
mp = importlib.reload(mp)
raft = importlib.reload(raft)

raft_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

def runRAFT(fname_design, fname_turbine, fname_env):
    '''
    This the main function for running the raft model in standalone form, where inputs are contained in the specified input files.
    '''
    
    # open the design YAML file and parse it into a dictionary for passing to raft
    with open(fname_design) as file:
        design = yaml.load(file, Loader=yaml.FullLoader)
    
    print("Loading file: "+fname_design)
    print(f"'{design['name']}'")
    
    depth = float(design['mooring']['water_depth'])
    
    # now off potMod in the design dictionary to avoid BEM analysis
    for mi in design['platform']['members']:    mi['potMod'] = False

    # set up frequency range
    w = np.arange(0.05, 5, 0.05)  # frequency range (to be set by modeling options yaml)
    
    # read in turbine data and combine it in
    # turbine = loadTurbineYAML(fname_turbine)
    # design['turbine'].update(turbine)
    
    
    # --- Create and run the model ---

    model = raft.Model(design, w=w, depth=depth)  # set up model
    
    model.setEnv(Hs=8, Tp=12, V=10, Fthrust=float(design['turbine']['Fthrust']))  # set basic wave and wind info

    model.calcSystemProps()          # get all the setup calculations done within the model
    
    model.solveEigen()
    
    model.calcMooringAndOffsets()    # calculate the offsets for the given loading
    
    model.solveDynamics()            # put everything together and iteratively solve the dynamic response

    model.plot()
    
    plt.show()
    
    return model
    

def loadTurbineYAML(fname_turbine):
    '''
    This loads data from a standard turbine YAML file to fill in the turbine portion of the RAFT input dictionary.
    '''
    
    # Set discretization parameters
    n_span = 30 # [-] - number of blade stations along span
    grid = np.linspace(0., 1., n_span) # equally spaced grid along blade span, root=0 tip=1
    n_aoa = 200 # [-] - number of angles of attack to discretize airfoil polars
    
    # dictionary to be filled in with turbine data
    d = dict(blade={}, airfoils={}, env={})
    
    # Load wind turbine geometry yaml
    print("Loading turbine YAML file: "+fname_turbine)
    
    run_dir = os.path.dirname( os.path.realpath(__file__) ) + os.sep
    fname_input_wt = os.path.join(run_dir, fname_turbine)
    wt_init = sch.load_geometry_yaml(fname_input_wt)
        
    print(f"'{wt_init['name']}'")
    
    # Conversion of the yaml inputs into CCBlade inputs
    Rhub = 0.5 * wt_init["components"]["hub"]["diameter"] # [m] - hub radius
    d['precone'] = np.rad2deg(wt_init["components"]["hub"]["cone_angle"]) # [deg] - rotor precone angle
    d['shaft_tilt'] = np.rad2deg(wt_init["components"]["nacelle"]["drivetrain"]["uptilt"]) # [deg] -  nacelle uptilt angle
    d['nBlades'] = wt_init["assembly"]["number_of_blades"] # [-] - number of blades

    # Blade quantities
    blade = wt_init["components"]["blade"]["outer_shape_bem"]
    rotor_diameter = wt_init["assembly"]["rotor_diameter"]
    blade_ref_axis = np.zeros((n_span, 3))
    blade_ref_axis[:, 0] = np.interp(grid, blade["reference_axis"]["x"]["grid"], blade["reference_axis"]["x"]["values"])
    blade_ref_axis[:, 1] = np.interp(grid, blade["reference_axis"]["y"]["grid"], blade["reference_axis"]["y"]["values"])
    blade_ref_axis[:, 2] = np.interp(grid, blade["reference_axis"]["z"]["grid"], blade["reference_axis"]["z"]["values"])
    if rotor_diameter != 0.0:
        blade_ref_axis[:, 2] = (blade_ref_axis[:, 2] * rotor_diameter / ((arc_length(blade_ref_axis)[-1] + Rhub) * 2.0))
    d['blade']['r'          ] = blade_ref_axis[1:-1, 2] + Rhub # [m] - radial position along straight blade pitch axis
    d['blade']['Rtip'       ] = blade_ref_axis[-1, 2] + Rhub
    d['blade']['chord'      ] = np.interp(grid[1:-1], blade["chord"]["grid"], blade["chord"]["values"]) # [m] - blade chord distributed along r
    d['blade']['theta'      ] = np.rad2deg(np.interp(grid[1:-1], blade["twist"]["grid"], blade["twist"]["values"])) # [deg] - blade twist distributed along r
    d['blade']['precurve'   ] = blade_ref_axis[1:-1, 0] # [m] - blade prebend distributed along r, usually negative for upwind rotors
    d['blade']['precurveTip'] = blade_ref_axis[-1, 0] # [m] - prebend at blade tip
    d['blade']['presweep'   ] = blade_ref_axis[1:-1, 1] # [m] - blade presweep distributed along r, usually positive
    d['blade']['presweepTip'] = blade_ref_axis[-1, 1] # [m] - presweep at blade tip

    # Hub height
    if wt_init["assembly"]["hub_height"] != 0.0:
        d['Zhub'] = wt_init["assembly"]["hub_height"]
    else:
        d['Zhub'] = wt_init["components"]["tower"]["outer_shape_bem"]["reference_axis"]["z"]["values"][-1] + wt_init["components"]["nacelle"]["drivetrain"]["distance_tt_hub"]

    # Atmospheric boundary layer data
    d['env']['rho'     ] = wt_init['environment']["air_density"] # [kg/m3] - density of air
    d['env']['mu'      ] = wt_init['environment']["air_dyn_viscosity"] # [kg/(ms)] - dynamic viscosity of air
    d['env']['shearExp'] = wt_init['environment']["shear_exp"] # [-] - shear exponent

    # Airfoil data
    n_af = len(wt_init["airfoils"])
    af_used = blade["airfoil_position"]["labels"]
    af_position = blade["airfoil_position"]["grid"]
    n_af_span = len(af_used)
    if n_aoa / 4.0 == int(n_aoa / 4.0):
        # One fourth of the angles of attack from -pi to -pi/6, half between -pi/6 to pi/6, and one fourth from pi/6 to pi
        aoa = np.unique(np.hstack([np.linspace(-np.pi, -np.pi / 6.0, int(n_aoa / 4.0 + 1)),np.linspace(-np.pi / 6.0,np.pi / 6.0,int(n_aoa / 2.0),),np.linspace(np.pi / 6.0, np.pi, int(n_aoa / 4.0 + 1))]))
    else:
        aoa = np.linspace(-np.pi, np.pi, n_aoa)
        print(
            "WARNING: If you like a grid of angles of attack more refined between +- 30 deg, please choose a n_aoa in the analysis option input file that is a multiple of 4. The current value of "
            + str(n_aoa)
            + " is not a multiple of 4 and an equally spaced grid is adopted."
        )

    Re_all = []
    for i in range(n_af):
        for j in range(len(wt_init["airfoils"][i]["polars"])):
            Re_all.append(wt_init["airfoils"][i]["polars"][j]["re"])
    n_Re = len(np.unique(Re_all))

    n_tab = 1

    af_name = n_af * [""]
    r_thick = np.zeros(n_af)
    Re_all = []
    for i in range(n_af):
        af_name[i] = wt_init["airfoils"][i]["name"]
        r_thick[i] = wt_init["airfoils"][i]["relative_thickness"]
        for j in range(len(wt_init["airfoils"][i]["polars"])):
            Re_all.append(wt_init["airfoils"][i]["polars"][j]["re"])

    Re = np.array(sorted(np.unique(Re_all)))

    cl = np.zeros((n_af, n_aoa, n_Re, n_tab))
    cd = np.zeros((n_af, n_aoa, n_Re, n_tab))
    cm = np.zeros((n_af, n_aoa, n_Re, n_tab))

    # Interp cl-cd-cm along predefined grid of angle of attack
    for i in range(n_af):
        n_Re_i = len(wt_init["airfoils"][i]["polars"])
        Re_j = np.zeros(n_Re_i)
        j_Re = np.zeros(n_Re_i, dtype=int)
        for j in range(n_Re_i):
            Re_j[j] = wt_init["airfoils"][i]["polars"][j]["re"]
            j_Re[j] = np.argmin(abs(Re - Re_j[j]))
            cl[i, :, j_Re[j], 0] = np.interp(
                aoa, wt_init["airfoils"][i]["polars"][j]["c_l"]["grid"], wt_init["airfoils"][i]["polars"][j]["c_l"]["values"]
            )
            cd[i, :, j_Re[j], 0] = np.interp(
                aoa, wt_init["airfoils"][i]["polars"][j]["c_d"]["grid"], wt_init["airfoils"][i]["polars"][j]["c_d"]["values"]
            )
            cm[i, :, j_Re[j], 0] = np.interp(
                aoa, wt_init["airfoils"][i]["polars"][j]["c_m"]["grid"], wt_init["airfoils"][i]["polars"][j]["c_m"]["values"]
            )

            if abs(cl[i, 0, j, 0] - cl[i, -1, j, 0]) > 1.0e-5:
                cl[i, 0, j, 0] = cl[i, -1, j, 0]
                print(
                    "WARNING: Airfoil "
                    + af_name[i]
                    + " has the lift coefficient at Re "
                    + str(Re_j[j])
                    + " different between + and - pi rad. This is fixed automatically, but please check the input data."
                )
            if abs(cd[i, 0, j, 0] - cd[i, -1, j, 0]) > 1.0e-5:
                cd[i, 0, j, 0] = cd[i, -1, j, 0]
                print(
                    "WARNING: Airfoil "
                    + af_name[i]
                    + " has the drag coefficient at Re "
                    + str(Re_j[j])
                    + " different between + and - pi rad. This is fixed automatically, but please check the input data."
                )
            if abs(cm[i, 0, j, 0] - cm[i, -1, j, 0]) > 1.0e-5:
                cm[i, 0, j, 0] = cm[i, -1, j, 0]
                print(
                    "WARNING: Airfoil "
                    + af_name[i]
                    + " has the moment coefficient at Re "
                    + str(Re_j[j])
                    + " different between + and - pi rad. This is fixed automatically, but please check the input data."
                )

        # Re-interpolate cl-cd-cm along the Re dimension if less than n_Re were provided in the input yaml (common condition)
        for l in range(n_aoa):
            cl[i, l, :, 0] = np.interp(Re, Re_j, cl[i, l, j_Re, 0])
            cd[i, l, :, 0] = np.interp(Re, Re_j, cd[i, l, j_Re, 0])
            cm[i, l, :, 0] = np.interp(Re, Re_j, cm[i, l, j_Re, 0])

    # Interpolate along blade span using a pchip on relative thickness
    r_thick_used = np.zeros(n_af_span)
    cl_used = np.zeros((n_af_span, n_aoa, n_Re, n_tab))
    #cl_interp = np.zeros((n_span, n_aoa, n_Re, n_tab))
    cd_used = np.zeros((n_af_span, n_aoa, n_Re, n_tab))
    #cd_interp = np.zeros((n_span, n_aoa, n_Re, n_tab))
    cm_used = np.zeros((n_af_span, n_aoa, n_Re, n_tab))
    #cm_interp = np.zeros((n_span, n_aoa, n_Re, n_tab))

    for i in range(n_af_span):
        for j in range(n_af):
            if af_used[i] == af_name[j]:
                r_thick_used[i] = r_thick[j]
                cl_used[i, :, :, :] = cl[j, :, :, :]
                cd_used[i, :, :, :] = cd[j, :, :, :]
                cm_used[i, :, :, :] = cm[j, :, :, :]
                break

    # Pchip does have an associated derivative method built-in:
    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.PchipInterpolator.derivative.html#scipy.interpolate.PchipInterpolator.derivative
    spline = PchipInterpolator
    rthick_spline = spline(af_position, r_thick_used)
    r_thick_interp = rthick_spline(grid[1:-1])

    # Spanwise interpolation of the airfoil polars with a pchip
    r_thick_unique, indices = np.unique(r_thick_used, return_index=True)
    cl_spline = spline(r_thick_unique, cl_used[indices, :, :, :])
    cl_interp = np.flip(cl_spline(np.flip(r_thick_interp)), axis=0)
    cd_spline = spline(r_thick_unique, cd_used[indices, :, :, :])
    cd_interp = np.flip(cd_spline(np.flip(r_thick_interp)), axis=0)
    cm_spline = spline(r_thick_unique, cm_used[indices, :, :, :])
    cm_interp = np.flip(cm_spline(np.flip(r_thick_interp)), axis=0)
    
    
    d['airfoils']['Re']        = Re
    d['airfoils']['aoa']       = aoa
    d['airfoils']['cl_interp'] = cl_interp
    d['airfoils']['cd_interp'] = cd_interp
    d['airfoils']['cm_interp'] = cm_interp
    
    d['Rhub'] = Rhub
    
    return d

    
    
def runRAFTfromWEIS():    
    ''' this is the more realistic case where we have to process wt_opt to produce memberStrings and MooringSystem <<<<<<<'''
        
    # Members
    floating_init_options = modeling_options['floating']  # I need to include these because this is where we get name_member
    n_members = floating_init_options['members']['n_members'] 
    
    n_joints = len(wt_opt['floating.floating_joints.location'])
    rA = np.zeros([n_joints, 2])
    rB = np.zeros([n_joints, 2])
    for i in range(n_joints):
        joint_locs[i,:] = wt_opt['floating.floating_joints.location'][i,:]
    
    for i in range(n_members):
        name_member = floating_init_options['members']['name'][i]
        type = 2 # arbitrary value to designate that the member is part of the floating substructure
        
        dA = wt_opt['floating.floating_member_' + name_member + '.outer_diameter'][0]
        dB = wt_opt['floating.floating_member_' + name_member + '.outer_diameter'][1]
        # <<<<<<<< the IEA ontology paper says that the outer_diameter parameter describes two diameters at joints 1 and 2
        
        t = sum(wt_opt['floating.floating_member_' + name_member + '.layer_thickness'])
        # took the sum of this because we just want to know the total thickness to get to dB
        # <<<<<<<<< not sure if I summed it right because the thickness of each layer is [j,:] in gc_WT_InitModel
        
        if n_joints != n_members + 1:
            raise ValueError('There should be n_members+1 number of joints to use the right rA and rB values')
        rA = joint_locs[i,:]
        rB = joint_locs[i+1,:]
        
        # <<<<<<<<<<< Ballast section: PROBABLY WON'T WORK. JUST USING WHAT I WAS GIVEN
        v_fill = wt_opt['floating.floating_member_' + name_member + '.ballast_volume'] 
        rho_fill = wt_opt['floating.floating_member_' + name_member + '.ballast_material.rho']

        #dB_fill = (dBi-dAi)*(self.l_fill/self.l) + dAi       # interpolated diameter of member where the ballast is filled to
        #v_fill = (np.pi/4)*(1/3)*(dAi**2+dB_fill**2+dAi*dB_fill)*self.l_fill    #[m^3]
        # There's a way to solve for l_fill using the above equations given v_fill
        
        # Going to simplify and just take it as the proportion of length to volume
        dAi = dA - 2*t # assming the thickness is constant along the member with respect to the length
        dBi = dB - 2*t
        l = np.linalg.norm(rB-rA)
        v_mem = (np.pi/4)*(1/3)*(dAi**2+dBi**2+dAi*dBi)*l
        
        l_fill = l * v_fill/v_mem
        
        # plug variables into a Member in FrequencyDomain and append to the memberString list
                    # change to string in FD v
        memberStrings.append(raft.Member( str(name_member)+" "+str(type)+" "+str(dA)+" "+str(dB)+" "+str(rA[0])+" "+str(rA[1])+" "+str(rA[2])+\
                                 " "+str(rB[0])+" "+str(rB[1])+" "+str(rB[2])+" "+str(t)+" "+str(l_fill)+" "+str(rho_fill), nw))
  
  
    
    # Create a MoorPy system
    ms = mp.System()
    ms.depth = wt_opt['env.water_depth']
    
    # Add the line types that are provided in the wt_opt OpenMDAO object
    n_line_types = len(wt_opt['mooring.line_diameter'])
    for i in range(n_line_types):
        name = wt_opt['mooring.line_names'][i]
        d = wt_opt['mooring.line_diameter'][i]
        massden = wt_opt['mooring.line_mass_density'][i]
        EA = wt_opt['mooring.line_stiffness'][i]
        MBL = wt_opt['mooring.line_breaking_load'][i]
        cost = wt_opt['mooring.line_cost'][i]
        
        ms.LineTypes[name] = mp.LineType( name, d, massden, EA, MBL=MBL, cost=cost, notes="made in FrequencyDomain.py" )
        
    # Add the wind turbine platform reference point   <<<<<<<<<<<<<< Get values
    ms.addBody(0, PRP, m=mTOT, v=VTOT, rCG=rCG_TOT, AWP=AWP_TOT, rM=np.array([0,0,zMeta]), f6Ext=np.array([Fthrust,0,0, 0,Mthrust,0]))
    
    # Add points to the sytem
    for i in range(n_nodes):
        ID = wt_opt['mooring.node_id'][i]            # <<<<<<<<< not 100% on the syntax of these calls
        
        if wt_opt['mooring.node_type'][i] == 'fixed':
            type = 1
        elif wt_opt['mooring.node_type'][i] == 'vessel':
            type = -1
        elif wt_opt['mooring.node_type'][i] == 'connection':
            type = 0
        
        r = np.array( wt_opt['mooring.nodes_location'][i,:], dtype=float)
        # TODO - can add in other variables for the point like anchor ID, fairlead_type, node_mass, node_volume, drag area, added mass
        ms.PointList.append( mp.Point( ID, type, r ) )

        # attach body points to the body
        # the nodes_location array is relative to inertial frame if Fixed or Connect, but relative to platform frame if Vessel
        if type==-1:
            ms.BodyList[0].addPoint(ID, r)
        
    
    # Add and attach lines to the nodes of the system
    n_lines = len(wt_opt['mooring.unstretched_length'])
    for i in range(n_lines):
        ID = wt_opt['mooring.line_id'][i]
        LineLength = wt_opt['mooring.unstretched_length'][i]
        linetype = wt_opt['mooring.line_type'][i]
        
        ms.LineList.append( mp.Line( ID, LineLength, LineTypes[linetype] ) )
        
        node1 = wt_opt['mooring.node1_id']
        node2 = wt_opt['mooring.node2_id']
        # Run an if statement to make sure that node1 is the deeper point
        if ms.PointList[node1].r[2] < ms.PointList[node2].r[2]:
            pass
        elif ms.PointList[node1].r[2] > ms.PointList[node2].r[2]:
            node1 = node2
            node2 = node1
        else:
            pass # if the z value of both points is the same, then it doesn't matter
        
        ms.PointList[node1].addLine(ID, 0)
        ms.PointList[node2].addLine(ID, 1)
    
        # TODO - anchor types
        
        # Turn on the system
        ms.initialize()
        MooringSystem = ms

    # NEED TO ADD THE FINAL MODEL RUN STEPS HERE ONCE THE ABOVE WORKS

    

if __name__ == "__main__":
    
    
    #model = runRAFT(os.path.join(raft_dir,'designs/OC3spar.yaml'), 'env.yaml')
    #model = runRAFT(os.path.join(raft_dir,'designs/OC4semi.yaml'), 'env.yaml')
    model = runRAFT(os.path.join(raft_dir,'designs/VolturnUS-S.yaml'),
                    os.path.join(raft_dir,'designs/rotors/IEA-15-240-RWT.yaml'),
                    os.path.join(raft_dir,'env.yaml'))
    #model = runRAFT(os.path.join(raft_dir,'designs/DTU10MW.yaml'),
    #                os.path.join(raft_dir,'designs/rotors/IEA-10-198-RWT.yaml'), 
    #                os.path.join(raft_dir,'env.yaml'))
    fowt = model.fowtList[0]


    
    
    
    """
    # ----- temporary script for comparing hydro coefficient curves -----

    # load the design
    with open('OC3spar.yaml') as file:
        design = yaml.load(file, Loader=yaml.FullLoader)
        
    depth = float(design['mooring']['water_depth'])
    w = np.arange(0.1, 5, 0.1)  # frequency range (to be set by modeling options yaml)
    
    # Create the model and compute hydrodynamic constants (with BEM)
    model1 = raft.Model(design, w=w, depth=depth)  # set up model
    model1.setEnv(spectrum="unit")  # set basic wave and wind info
    model1.calcSystemProps()          # get all the setup calculations done within the model
    
    # now turn off PotMod in the design dictionary
    for mi in design['platform']['members']:    mi['potMod'] = False
        
    # Create another model and compute hydrodynamic constants (with strip theory only)
    model2 = raft.Model(design, w=w, depth=depth)  # set up model
    model2.setEnv(spectrum='unit')
    model2.calcSystemProps()          # get all the setup calculations done within the model
    
    
    fix, ax = plt.subplots(3,1, sharex=True)
    for i, j in enumerate([0,2,4]):       # go through surge, heave, and pitch
        ax[i].plot(model1.w, model1.fowtList[0].F_BEM       [j].real, 'b'  , label="F BEM real")
        ax[i].plot(model1.w, model1.fowtList[0].F_BEM       [j].imag, 'b--', label="F BEM imag")
        ax[i].plot(model2.w, model2.fowtList[0].F_hydro_iner[j].real, 'g'  , label="F Froude-Kry real")
        ax[i].plot(model2.w, model2.fowtList[0].F_hydro_iner[j].imag, 'g--', label="F Froude-Kry imag")
    ax[-1].legend()
    
    ax[0].set_ylabel('surge')
    ax[1].set_ylabel('heave')
    ax[2].set_ylabel('pitch')
    
    
    # plot member distributed stuff now
    
    f = model2.fowtList[0]
    
    n = f.memberList[0].F_exc_iner.shape[0]
    n3 = int(np.ceil(n/3))
    
    fig, ax = plt.subplots(n3,3, sharex=True)
    for i in range(3):       # go through surge, heave, and pitch
        for j in range(n3):
        
            k = i*n3 + j
            
            if k < n:
                    
                ax[j,i].axhline(0.0, color='k', lw=0.4)
                ax[j,i].plot(model2.w, model2.fowtList[0].memberList[0].F_exc_iner[k,0,:].real, 'g'  , label="Fx real")
                ax[j,i].plot(model2.w, model2.fowtList[0].memberList[0].F_exc_iner[k,0,:].imag, 'g--', label="Fx imag")
                ax[j,i].plot(model2.w, model2.fowtList[0].memberList[0].F_exc_iner[k,2,:].real, 'r'  , label="Fz real")
                ax[j,i].plot(model2.w, model2.fowtList[0].memberList[0].F_exc_iner[k,2,:].imag, 'r--', label="Fz imag")
                ax[j,i].set_ylabel(f"strip {k}")
        
    ax[-1,-1].legend()
    
    
    plt.show()
    """
    



# tests RAFT FOWT functionality and results

import pytest
import numpy as np
from numpy.testing import assert_allclose
import yaml
import pickle
import raft
from raft.raft_rotor import Rotor
from raft.helpers import getFromDict
import os

'''
 Define files for testing
'''
# Name of the subfolder where the test data is located
test_dir = 'test_data'

# List of input file names to be tested
list_files = [
    'IEA15MW.yaml',
]

# Get the directory of the current file
current_dir = os.path.dirname(os.path.abspath(__file__))

# To avoid problems with different platforms, get the full path of the file
list_files = [os.path.join(current_dir, test_dir, file) for file in list_files]


'''
 Aux functions
'''
# Function used to create Rotor instance
# Not explicitly inside the fixture below so that we can also run this file as a script
# This code is a modified copy from parts of FOWT.__init__ and Model.__init__ 
# We should probably have a function for doing that to avoid code repetition.
def create_rotor(file):
    with open(file) as f:
        design = yaml.load(f, Loader=yaml.FullLoader)
        
    nrotors = getFromDict(design['turbine'], 'nrotors', dtype=int, shape=0, default=1)
    if nrotors==1: design['turbine']['nrotors'] = 1
    else: raise NotImplementedError('Multiple rotors not supported yet by the testing function.')
    
    if 'tower' in design['turbine']:
        if isinstance(design['turbine']['tower'], dict):                           # if a single tower is specified (rather than a list)
            design['turbine']['tower'] = [design['turbine']['tower']]*nrotors  # replicate the tower info for each rotor    

    # copy over site info into turbine dictionary
    design['turbine']['rho_air'       ] = getFromDict(design['site'], 'rho_air', shape=0, default=1.225)
    design['turbine']['mu_air'        ] = getFromDict(design['site'], 'mu_air', shape=0, default=1.81e-05)
    design['turbine']['shearExp_air'  ] = getFromDict(design['site'], 'shearExp_air', shape=0, default=0.12)
    design['turbine']['rho_water'     ] = getFromDict(design['site'], 'rho_water', shape=0, default=1025.0)
    design['turbine']['mu_water'      ] = getFromDict(design['site'], 'mu_water', shape=0, default=1.0e-03)
    design['turbine']['shearExp_water'] = getFromDict(design['site'], 'shearExp_water', shape=0, default=0.12)

    min_freq     = getFromDict(design['settings'], 'min_freq', default=0.01, dtype=float)  # [Hz] lowest frequency to consider, also the frequency bin width 
    max_freq     = getFromDict(design['settings'], 'max_freq', default=1.00, dtype=float)  # [Hz] highest frequency to consider        
    w = np.arange(min_freq, max_freq+0.5*min_freq, min_freq) *2*np.pi  # angular frequencies to analyze (rad/s)
    
    # Nacelle geometry (for hydro): if a single one (dictionary) is listed,
    # replicate it into a list for all rotors.
    if 'nacelle' in design['turbine']:
        if isinstance(design['turbine']['nacelle'], dict): 
            design['turbine']['nacelle'] = [design['turbine']['nacelle']]*nrotors

    return Rotor(design['turbine'], w, 0)


# Define a fixture to loop rotor instances with the index to loop the desired values as well
# Could also zip the lists with the desired values, but I think the approach below is simpler
@pytest.fixture(params=enumerate(list_files))
def index_and_rotor(request):
    index, file = request.param
    rotor = create_rotor(file)    
    return index, rotor


'''
 Test functions
'''
def test_calcAero(index_and_rotor, flagSaveValues=False):    
    '''
    Verify rotor.calcAero for many combinations of wind speed, heading, TI, and yaw mode.
    Some combinations seem to be outside the validity of CCBlade (e.g., yaw_mode=1, wind direction=90, and turbine_heading=0),
    but they are tested nonetheless.
    Set flagSaveValues to true to replace the true values file with the values calculated below
    '''
    index, rotor = index_and_rotor
    rotor.setPosition()

    if 'IEA15MW' in list_files[index]:
        U_rated = 10.59
    elif 'NREL5MW' in list_files[index]:
        U_rated = 11.4
    else:
        U_rated = 12
        print(f'Unknown turbine. Considering U_rated = {U_rated} m/s.')

    # We are going to loop through the following values to test a broad range of cases
    wind_speeds = np.sort(np.append([5, 15, 25], U_rated))
    wind_headings = [-45, 0, 45]
    TI            = [0, 0.5]
    yaw_modes     = [0, 1, 2, 3]
    values4control = [  # Values used for yaw misalignment or turbine heading depending on yaw_mode. One list per yaw_mode
                     [0],
                     [-15, 0, 15],
                     [-15, 0, 15],
                     [-15, 0, 15],
                     ]
    
    for idx_ym, ym in enumerate(yaw_modes):
        rotor.yaw_mode = ym
    
        true_values_file = list_files[index].replace('.yaml', f'_true_calcAero-yaw_mode{ym:d}.pkl')
        output_true_values = []
        idxTrueValues = 0

        for ws in wind_speeds:
            for wh in wind_headings:
                for ti in TI:
                    for v4c in values4control[idx_ym]:
                        if ym == 1: # This one needs turbine_heading
                            thisCase = {'wind_speed': ws, 'wind_heading': wh, 'turbulence': ti, 'turbine_status': 'operating', 'turbine_heading': v4c}
                        else:
                            thisCase = {'wind_speed': ws, 'wind_heading': wh, 'turbulence': ti, 'turbine_status': 'operating', 'yaw_misalign': v4c}
                        f_aero0, f_aero, a_aero, b_aero = rotor.calcAero(thisCase)

                        if flagSaveValues:
                            output_true_values.append({
                                'case': thisCase,
                                'f_aero0': f_aero0,
                                'f_aero': f_aero,
                                'a_aero': a_aero,
                                'b_aero': b_aero
                            })

                        else:
                            with open(true_values_file, 'rb') as f:
                                true_values = pickle.load(f)

                            assert_allclose(f_aero0, true_values[idxTrueValues]['f_aero0'], rtol=1e-5, atol=1e-5)
                            assert_allclose(f_aero, true_values[idxTrueValues]['f_aero'], rtol=1e-5, atol=1e-5)
                            assert_allclose(a_aero, true_values[idxTrueValues]['a_aero'], rtol=1e-5, atol=1e-5)
                            assert_allclose(b_aero, true_values[idxTrueValues]['b_aero'], rtol=1e-5, atol=1e-5)
                        idxTrueValues += 1

            if flagSaveValues:
                with open(true_values_file, 'wb') as f:
                    pickle.dump(output_true_values, f)
    # Do something similar for turbine class

'''
 To run as a script. Useful for debugging.
'''
if __name__ == "__main__":
    index = 0

    rotor = create_rotor(list_files[index])
    test_calcAero((index, rotor), flagSaveValues=False)




# This is a file to handle meshing of things in RAFT for the time being. It outputs in the HAMS .pnl format.


import numpy as np
import os
import os.path as osp

def makePanel(X, Y, Z, savedNodes, savedPanels):
    '''
    Sets up panel and node data for HAMS .pnl input file.
    Also ensures things don't go above the water line. (A rough implementation that should be improved.)

    X, Y, Z : lists
        panel coordinates - 4 expected
    savedNodes : list of lists
        all the node coordinates already saved
    savedPanels : list
        the information for all the panels already saved: panel_number number_of_vertices   Vertex1_ID   Vertex2_ID  Vertex3_ID   (Vertex4_ID)

    '''
    
    
    # if the panel is fully out of the water, skip it
    if (np.array(Z) > 0.0).all():   
        return
    
    # if any points are above the water, bring them down to the water surface
    for i in range(4):
        if Z[i] > 0.0:
            Z[i] = 0.0
            
    # now process the node points w.r.t. existing list
    #points = np.vstack([X, Y, Z])                # make a single 2D array for easy manipulation
    
    pNodeIDs = []                                    # the indices of the nodes for this panel (starts at 1)
    
    pNumNodes = 4
    
    for i in range(4):
    
        ndi = [X[i],Y[i],Z[i]]                       # the current node in question in the panel
        
        match = [j+1 for j,nd in enumerate(savedNodes) if nd==ndi]  # could do this in reverse order for speed...
        
        if len(match)==0:                            # point does not already exist in list; add it.
            savedNodes.append(ndi)
            pNodeIDs.append(len(savedNodes))
            
        elif len(match)==1:                          # point exists in list; refer to it rather than adding another point
            
            if match[0] in pNodeIDs:                 # if the current panel has already referenced this node index, convert to tri
                #print("triangular panel detected!")  # we will skip adding this point to the panel's indix list                                                     
                pNumNodes -= 1                       # reduce the number of nodes for this panel
                
            else:
                pNodeIDs.append(match[0])            # otherwise add this point index to the panel's list like usual
            
        else:
            ValueError("Somehow there are duplicate points in the list!")
            
    panelID = len(savedPanels)+1                     # id number of the current panel (starts from 1)
    
    
    if pNumNodes == 4:
        savedPanels.append([panelID, pNumNodes, pNodeIDs[0], pNodeIDs[1], pNodeIDs[2], pNodeIDs[3]])
    elif pNumNodes == 3:
        savedPanels.append([panelID, pNumNodes, pNodeIDs[0], pNodeIDs[1], pNodeIDs[2]])
    else:
        ValueError(f"Somehow there are only {pNumNodes} unique nodes for panel {panelID}")



def meshMember(stations, diameters, rA, rB, dz_max=0, da_max=0, savedNodes=[], savedPanels=[]):
    '''
    Creates mesh for an axisymmetric member as defined by RAFT.

    Parameters
    ----------
    stations:  list 
        locations along member axis at which the cross section will be specified
    diameters: list 
        corresponding diameters along member
    rA, rB: list
        member end point coordinates
    dz_max: float
        maximum panel height
    da_max: float
        maximum panel width (before doubling azimuthal discretization)
    savedNodes : list of lists
        all the node coordinates already saved
    savedPanels : list
        the information for all the panels already saved: panel_number number_of_vertices   Vertex1_ID   Vertex2_ID  Vertex3_ID   (Vertex4_ID)


    Returns
    -------
    nodes : list
        list of node coordinates
    panels : list
        the information for all the panels: panel_number number_of_vertices   Vertex1_ID   Vertex2_ID  Vertex3_ID   (Vertex4_ID)

    '''

    
    radii = 0.5*np.array(diameters)


    # discretization defaults
    if dz_max==0:
        dz_max = stations[-1]/20
    if da_max==0:
        da_max = np.max(radii)/8
        

    # ------------------ discretize radius profile according to dz_max --------

    # radius profile data is contained in r_rp and z_rp
    r_rp = [radii[0]]
    z_rp = [0.0]

    # step through each station and subdivide as needed
    for i_s in range(1, len(radii)):
        dr_s = radii[i_s] - radii[i_s-1]; # delta r
        dz_s = stations[ i_s] - stations[ i_s-1]; # delta z
        # subdivision size
        if dr_s == 0: # vertical case
            cos_m=1
            sin_m=0
            dz_ps = dz_max; # (dz_ps is longitudinal dimension of panel)
        elif dz_s == 0: # horizontal case
            cos_m=0
            sin_m=np.sign(dr_s)
            dz_ps = 0.6*da_max
        else: # angled case - set panel size as weighted average based on slope
            m = dr_s/dz_s; # slope = dr/dz
            dz_ps = np.arctan(np.abs(m))*2/np.pi*0.6*da_max + np.arctan(abs(1/m))*2/np.pi*dz_max;
            cos_m = dz_s/np.sqrt(dr_s**2 + dz_s**2)
            sin_m = dr_s/np.sqrt(dr_s**2 + dz_s**2)
        # make subdivision
        # local panel longitudinal discretization
        n_z = int(np.ceil( np.sqrt(dr_s*dr_s + dz_s*dz_s) / dz_ps ))

        # local panel longitudinal dimension
        d_l = np.sqrt(dr_s*dr_s + dz_s*dz_s)/n_z;
        for i_z in range(1,n_z+1):
            r_rp.append(  radii[i_s-1] + sin_m*i_z*d_l)
            z_rp.append(stations[i_s-1] + cos_m*i_z*d_l)
            

    # fill in end B if it's submerged
    n_r = int(np.ceil( radii[-1] / (0.6*da_max) ))   # local panel radial discretization #
    dr  = radii[-1] / n_r                               # local panel radial size

    for i_r in range(n_r):
        r_rp.append(radii[-1] - (1+i_r)*dr)
        z_rp.append(stations[-1])
    
    
    # fill in end A if it's submerged
    n_r = int(np.ceil( radii[0] / (0.6*da_max) ))   # local panel radial discretization #
    dr  = radii[0] / n_r                               # local panel radial size

    for i_r in range(n_r):
        r_rp.insert(0, radii[0] - (1+i_r)*dr)
        z_rp.insert(0, stations[0])
    
    
    # --------------- revolve radius profile, do adaptive paneling stuff ------

    # lists that we'll put the panel coordinates in, in lists of 4 coordinates per panel
    x = []
    y = []
    z = []

    npan =0;
    naz = int(8);

    # go through each point of the radius profile, panelizing from top to bottom:
    for i_rp in range(len(z_rp)-1):
    
        # rectangle coords - shape from outside is:  A D
        #                                            B C
        r1=r_rp[i_rp];
        r2=r_rp[i_rp+1];
        z1=z_rp[i_rp];
        z2=z_rp[i_rp+1];

        # scale up or down azimuthal discretization as needed
        while ( (r1*2*np.pi/naz >= da_max/2) and (r2*2*np.pi/naz >= da_max/2) ):
            naz = int(2*naz)
        while ( (r1*2*np.pi/naz < da_max/2) and (r2*2*np.pi/naz < da_max/2) ):
            naz = int(naz/2)

        # transition - increase azimuthal discretization
        if ( (r1*2*np.pi/naz < da_max/2) and (r2*2*np.pi/naz >= da_max/2) ):
            for ia in range(1, int(naz/2)+1):
                th1 = (ia-1  )*2*np.pi/naz*2
                th2 = (ia-0.5)*2*np.pi/naz*2
                th3 = (ia    )*2*np.pi/naz*2

                x.append([(r1*np.cos(th1)+r1*np.cos(th3))/2, r2*np.cos(th2), r2*np.cos(th1), r1*np.cos(th1)])
                y.append([(r1*np.sin(th1)+r1*np.sin(th3))/2, r2*np.sin(th2), r2*np.sin(th1), r1*np.sin(th1)])
                z.append([z1                               , z2            , z2            , z1            ])

                npan += 1

                x.append([r1*np.cos(th3), r2*np.cos(th3), r2*np.cos(th2), (r1*np.cos(th1)+r1*np.cos(th3))/2])
                y.append([r1*np.sin(th3), r2*np.sin(th3), r2*np.sin(th2), (r1*np.sin(th1)+r1*np.sin(th3))/2])
                z.append([z1            , z2            , z2            , z1                               ])

                npan += 1

        # transition - decrease azimuthal discretization
        elif ( (r1*2*np.pi/naz >= da_max/2) and (r2*2*np.pi/naz < da_max/2) ):
            for ia in range(1, int(naz/2)+1):
                th1 = (ia-1  )*2*np.pi/naz*2
                th2 = (ia-0.5)*2*np.pi/naz*2
                th3 = (ia    )*2*np.pi/naz*2

                x.append([r1*np.cos(th2), r2*(np.cos(th1)+np.cos(th3))/2, r2*np.cos(th1), r1*np.cos(th1)])
                y.append([r1*np.sin(th2), r2*(np.sin(th1)+np.sin(th3))/2, r2*np.sin(th1), r1*np.sin(th1)])
                z.append([z1            , z2                            , z2            , z1            ])

                npan += 1;

                x.append([r1*np.cos(th3), r2*np.cos(th3), r2*(np.cos(th1)+np.cos(th3))/2, r1*np.cos(th2)])
                y.append([r1*np.sin(th3), r2*np.sin(th3), r2*(np.sin(th1)+np.sin(th3))/2, r1*np.sin(th2)])
                z.append([z1            , z2            , z2                            , z1            ])

                npan += 1

        # no transition
        else:
            for ia in range(1, naz+1):
                th1 = (ia-1)*2*np.pi/naz
                th2 = (ia  )*2*np.pi/naz
                
                x.append([r1*np.cos(th2), r2*np.cos(th2), r2*np.cos(th1), r1*np.cos(th1)])
                y.append([r1*np.sin(th2), r2*np.sin(th2), r2*np.sin(th1), r1*np.sin(th1)])
                z.append([z1            , z2            , z2            , z1            ])

                npan += 1



    # calculate member rotation matrix 
    rAB = rB - rA                                               # displacement vector from end A to end B [m]
    
    beta = np.arctan2(rAB[1],rAB[0])                            # member incline heading from x axis
    phi  = np.arctan2(np.sqrt(rAB[0]**2 + rAB[1]**2), rAB[2])   # member incline angle from vertical
    
    s1 = np.sin(beta) 
    c1 = np.cos(beta)
    s2 = np.sin(phi) 
    c2 = np.cos(phi)
    s3 = np.sin(0.0) 
    c3 = np.cos(0.0)

    R = np.array([[ c1*c2*c3-s1*s3, -c3*s1-c1*c2*s3,  c1*s2],
                  [ c1*s3+c2*c3*s1,  c1*c3-c2*s1*s3,  s1*s2],
                  [   -c3*s2      ,      s2*s3     ,    c2 ]])  #Z1Y2Z3 from https://en.wikipedia.org/wiki/Euler_angles#Rotation_matrix
    
    nSavedPanelsOld = len(savedPanels)
    
    # adjust each panel position based on member pose then set up the member
    for i in range(npan):    
    
        nodes0 = np.array([x[i], y[i], z[i]])           # the nodes of this panel
        
        nodes = np.matmul( R, nodes0 ) + rA[:,None]    # rotate and translate nodes of this panel based on member pose
    
        makePanel(nodes[0], nodes[1], nodes[2], savedNodes, savedPanels)

    print(f'Of {npan} generated panels, {len(savedPanels)-nSavedPanelsOld} were submerged and have been used in the mesh.')

    return savedNodes, savedPanels
    
    

def writeMesh(savedNodes, savedPanels, oDir=""):
    '''Creates a HAMS .pnl file based on savedNodes and savedPanels lists'''
        
    numPanels   = len(savedPanels)
    numNodes    = len(savedNodes)    
        
    # write .pnl file
    if osp.isdir(oDir) is not True:
        os.makedirs(oDir)
        
    oFilePath = os.path.join(oDir, 'HullMesh.pnl')

    oFile = open(oFilePath, 'w')
    oFile.write('    --------------Hull Mesh File---------------\n\n')
    oFile.write('    # Number of Panels, Nodes, X-Symmetry and Y-Symmetry\n')
    oFile.write(f'         {numPanels}         {numNodes}         0         0\n\n')

    oFile.write('    #Start Definition of Node Coordinates     ! node_number   x   y   z\n')
    for i in range(numNodes):
        oFile.write(f'{i+1:>5}{savedNodes[i][0]:18.3f}{savedNodes[i][1]:18.3f}{savedNodes[i][2]:18.3f}\n')
    oFile.write('   #End Definition of Node Coordinates\n\n')
    
    oFile.write('   #Start Definition of Node Relations   ! panel_number  number_of_vertices   Vertex1_ID   Vertex2_ID   Vertex3_ID   (Vertex4_ID)\n')
    for i in range(numPanels):
        oFile.write(''.join([f'{p:>8}' for p in savedPanels[i]])+'\n')
    oFile.write('   #End Definition of Node Relations\n\n')
    oFile.write('    --------------End Hull Mesh File---------------\n')
    
    oFile.close()
    

# below are some helper functions for outputting in GDF format instead, for visualization at this point


def meshMemberForGDF(stations, diameters, rA, rB, dz_max=0, da_max=0, endA=True, endB=True):
    '''
    Creates mesh for an axisymmetric member as defined by RAFT.

    Parameters
    ----------
    stations:  list of locations along member axis at which the cross section will be specified
    diameters: list of corresponding diameters along member
    rA, rB: member end point coordinates
    dz_max: maximum panel height
    da_max: maximum panel width (before doubling azimuthal discretization)
    endA/endB: flag for whether to mesh each end (may want to add for meshMember too)

    Returns
    -------
    vertices : array
        An array containing the mesh point coordinates, size [3, 4*npanel]
    '''

    
    radii = 0.5*np.array(diameters)


    # discretization defaults
    if dz_max==0:
        dz_max = stations[-1]/20
    if da_max==0:
        da_max = np.max(radii)/8
        

    # ------------------ discretize radius profile according to dz_max --------

    # radius profile data is contained in r_rp and z_rp
    r_rp = [radii[0]]
    z_rp = [0.0]

    # step through each station and subdivide as needed
    for i_s in range(1, len(radii)):
        dr_s = radii[i_s] - radii[i_s-1]; # delta r
        dz_s = stations[ i_s] - stations[ i_s-1]; # delta z
        # subdivision size
        if dr_s == 0: # vertical case
            cos_m=1
            sin_m=0
            dz_ps = dz_max; # (dz_ps is longitudinal dimension of panel)
        elif dz_s == 0: # horizontal case
            cos_m=0
            sin_m=np.sign(dr_s)
            dz_ps = 0.6*da_max
        else: # angled case - set panel size as weighted average based on slope
            m = dr_s/dz_s; # slope = dr/dz
            dz_ps = np.arctan(np.abs(m))*2/np.pi*0.6*da_max + np.arctan(abs(1/m))*2/np.pi*dz_max;
            cos_m = dz_s/np.sqrt(dr_s**2 + dz_s**2)
            sin_m = dr_s/np.sqrt(dr_s**2 + dz_s**2)
        # make subdivision
        # local panel longitudinal discretization
        n_z = int(np.ceil( np.sqrt(dr_s*dr_s + dz_s*dz_s) / dz_ps ))
        # local panel longitudinal dimension
        d_l = np.sqrt(dr_s*dr_s + dz_s*dz_s)/n_z;
        for i_z in range(1,n_z+1):
            r_rp.append(  radii[i_s-1] + sin_m*i_z*d_l)
            z_rp.append(stations[i_s-1] + cos_m*i_z*d_l)
        

    # fill in end B if it's requested
    if endB:
        n_r = int(np.ceil( radii[-1] / (0.6*da_max) ))   # local panel radial discretization #
        dr  = radii[-1] / n_r                               # local panel radial size

        for i_r in range(n_r):
            r_rp.append(radii[-1] - (1+i_r)*dr)
            z_rp.append(stations[-1])
    
    
    # fill in end A if it's requested
    if endA:
        n_r = int(np.ceil( radii[0] / (0.6*da_max) ))   # local panel radial discretization #
        dr  = radii[0] / n_r                               # local panel radial size

        for i_r in range(n_r):
            r_rp.insert(0, radii[0] - (1+i_r)*dr)
            z_rp.insert(0, stations[0])
    
    
    # --------------- revolve radius profile, do adaptive paneling stuff ------

    # lists that we'll put the panel coordinates in, in lists of 4 coordinates per panel
    x = []
    y = []
    z = []

    npan =0;
    naz = int(8);

    # go through each point of the radius profile, panelizing from top to bottom:
    for i_rp in range(len(z_rp)-1):
    
        # rectangle coords - shape from outside is:  A D
        #                                            B C
        r1=r_rp[i_rp];
        r2=r_rp[i_rp+1];
        z1=z_rp[i_rp];
        z2=z_rp[i_rp+1];

        # scale up or down azimuthal discretization as needed
        while ( (r1*2*np.pi/naz >= da_max/2) and (r2*2*np.pi/naz >= da_max/2) ):
            naz = int(2*naz)
        while ( (r1*2*np.pi/naz < da_max/2) and (r2*2*np.pi/naz < da_max/2) ):
            naz = int(naz/2)

        # transition - increase azimuthal discretization
        if ( (r1*2*np.pi/naz < da_max/2) and (r2*2*np.pi/naz >= da_max/2) ):
            for ia in range(1, int(naz/2)+1):
                th1 = (ia-1  )*2*np.pi/naz*2;
                th2 = (ia-0.5)*2*np.pi/naz*2;
                th3 = (ia    )*2*np.pi/naz*2;

                x += [(r1*np.cos(th1)+r1*np.cos(th3))/2, r2*np.cos(th2), r2*np.cos(th1), r1*np.cos(th1) ]
                y += [(r1*np.sin(th1)+r1*np.sin(th3))/2, r2*np.sin(th2), r2*np.sin(th1), r1*np.sin(th1) ]
                z += [z1                               , z2            , z2            , z1             ]

                npan += 1

                x += [r1*np.cos(th3), r2*np.cos(th3), r2*np.cos(th2), (r1*np.cos(th1)+r1*np.cos(th3))/2]
                y += [r1*np.sin(th3), r2*np.sin(th3), r2*np.sin(th2), (r1*np.sin(th1)+r1*np.sin(th3))/2]
                z += [z1            , z2            , z2            , z1                               ]

                npan += 1

        # transition - decrease azimuthal discretization
        elif ( (r1*2*np.pi/naz >= da_max/2) and (r2*2*np.pi/naz < da_max/2) ):
            for ia in range(1, int(naz/2)+1):
                th1 = (ia-1  )*2*np.pi/naz*2;
                th2 = (ia-0.5)*2*np.pi/naz*2;
                th3 = (ia    )*2*np.pi/naz*2;
                x += [r1*np.cos(th2), r2*(np.cos(th1)+np.cos(th3))/2, r2*np.cos(th1), r1*np.cos(th1)]
                y += [r1*np.sin(th2), r2*(np.sin(th1)+np.sin(th3))/2, r2*np.sin(th1), r1*np.sin(th1)]
                z += [z1            , z2                            , z2            , z1            ]

                npan += 1;

                x += [r1*np.cos(th3), r2*np.cos(th3), r2*(np.cos(th1)+np.cos(th3))/2, r1*np.cos(th2)]
                y += [r1*np.sin(th3), r2*np.sin(th3), r2*(np.sin(th1)+np.sin(th3))/2, r1*np.sin(th2)]
                z += [z1            , z2            , z2                            , z1            ]

                npan += 1

        # no transition
        else:
            for ia in range(1, naz+1):
                th1 = (ia-1)*2*np.pi/naz;
                th2 = (ia  )*2*np.pi/naz;
                x += [r1*np.cos(th2), r2*np.cos(th2), r2*np.cos(th1), r1*np.cos(th1)]
                y += [r1*np.sin(th2), r2*np.sin(th2), r2*np.sin(th1), r1*np.sin(th1)]
                z += [z1            , z2            , z2            , z1            ]

                npan += 1

    # ----- transform coordinates to reflect specified rA and rB values
    
    vertices = np.array([x, y, z])

    rAB = rB - rA                                               # displacement vector from end A to end B [m]
    
    beta = np.arctan2(rAB[1],rAB[0])                            # member incline heading from x axis
    phi  = np.arctan2(np.sqrt(rAB[0]**2 + rAB[1]**2), rAB[2])   # member incline angle from vertical
    
    # trig terms for Euler angles rotation based on beta, phi, and gamma
    s1 = np.sin(beta) 
    c1 = np.cos(beta)
    s2 = np.sin(phi) 
    c2 = np.cos(phi)
    s3 = np.sin(0.0) 
    c3 = np.cos(0.0)

    R = np.array([[ c1*c2*c3-s1*s3, -c3*s1-c1*c2*s3,  c1*s2],
                  [ c1*s3+c2*c3*s1,  c1*c3-c2*s1*s3,  s1*s2],
                  [   -c3*s2      ,      s2*s3     ,    c2 ]])  #Z1Y2Z3 from https://en.wikipedia.org/wiki/Euler_angles#Rotation_matrix
    
    vertices2 = np.matmul( R, vertices ) + rA[:,None]
    
    
    # >>>> would need a step here to stop the mesh at the waterplane <<<<
    
    
    return vertices2.T


# For rectangular members

def meshRectangularMember(stations, widths, heights, rA, rB, dz_max=0, dw_max=0, dh_max=0, savedNodes=[], savedPanels=[]):
    '''
    Creates mesh for a rectangular member.

    Parameters
    ----------
    stations:  list 
        locations along member axis at which the cross section will be specified
    widths: list 
        corresponding widths along member
    heights: list
        corresponding heights along member
    rA, rB: list
        member end point coordinates
    dz_max: float
        maximum panel height
    dw_max: float
        maximum panel width
    dh_max: float
        maximum panel height
    savedNodes : list of lists
        all the node coordinates already saved
    savedPanels : list
        the information for all the panels already saved: panel_number number_of_vertices   Vertex1_ID   Vertex2_ID   Vertex3_ID   (Vertex4_ID)

    Returns
    -------
    nodes : list
        list of node coordinates
    panels : list
        the information for all the panels: panel_number number_of_vertices   Vertex1_ID   Vertex2_ID   Vertex3_ID   (Vertex4_ID)
    '''

    # discretization defaults
    if dz_max == 0:
        dz_max = stations[-1] / 20
    if dw_max == 0:
        dw_max = np.max(widths) / 8
    if dh_max == 0:
        dh_max = np.max(heights) / 8

    # calculate member rotation matrix
    rAB = rB - rA  # displacement vector from end A to end B [m]
    beta = np.arctan2(rAB[1], rAB[0])  # member incline heading from x axis
    phi = np.arctan2(np.sqrt(rAB[0]**2 + rAB[1]**2), rAB[2])  # member incline angle from vertical

    s1 = np.sin(beta)
    c1 = np.cos(beta)
    s2 = np.sin(phi)
    c2 = np.cos(phi)
    s3 = np.sin(0.0)
    c3 = np.cos(0.0)

    R = np.array([[c1 * c2 * c3 - s1 * s3, -c3 * s1 - c1 * c2 * s3, c1 * s2],
                  [c1 * s3 + c2 * c3 * s1, c1 * c3 - c2 * s1 * s3, s1 * s2],
                  [-c3 * s2, s2 * s3, c2]])  # Z1Y2Z3 from https://en.wikipedia.org/wiki/Euler_angles#Rotation_matrix

    nSavedPanelsOld = len(savedPanels)

    # Create intermediate stations based on dz_max
    new_stations = [stations[0]]
    new_widths = [widths[0]]
    new_heights = [heights[0]]
    for i in range(len(stations) - 1):
        z1, z2 = stations[i], stations[i + 1]
        w1, w2 = widths[i], widths[i + 1]
        h1, h2 = heights[i], heights[i + 1]

        dz = z2 - z1
        num_divisions = max(int(np.ceil(dz / dz_max)), 1)
        z_divisions = np.linspace(z1, z2, num_divisions + 1)
        for j in range(1, num_divisions + 1):
            new_stations.append(z_divisions[j])
            new_widths.append(w1 + (w2 - w1) * (z_divisions[j] - z1) / dz)
            new_heights.append(h1 + (h2 - h1) * (z_divisions[j] - z1) / dz)

    # Generate panels for all divisions
    for i in range(len(new_stations) - 1):
        w1, w2 = new_widths[i], new_widths[i + 1]
        h1, h2 = new_heights[i], new_heights[i + 1]
        z1, z2 = new_stations[i], new_stations[i + 1]

        nw = max(int(np.ceil(w1 / dw_max)), int(np.ceil(w2 / dw_max)))
        nh = max(int(np.ceil(h1 / dh_max)), int(np.ceil(h2 / dh_max)))

        x_divisions_w1 = np.linspace(-w1 / 2, w1 / 2, nw + 1)
        x_divisions_w2 = np.linspace(-w2 / 2, w2 / 2, nw + 1)
        y_divisions_h1 = np.linspace(-h1 / 2, h1 / 2, nh + 1)
        y_divisions_h2 = np.linspace(-h2 / 2, h2 / 2, nh + 1)

        for iw in range(nw):
            for ih in range(nh):
                # Panel 1: Front face
                X = [x_divisions_w1[iw], x_divisions_w1[iw + 1], x_divisions_w2[iw + 1], x_divisions_w2[iw]]
                Y = [-h1 / 2, -h1 / 2, -h2 / 2, -h2 / 2]
                Z = [z1, z1, z2, z2]
                nodes0 = np.array([X, Y, Z])
                nodes = np.matmul(R, nodes0) + rA[:, None]
                makePanel(nodes[0], nodes[1], nodes[2], savedNodes, savedPanels)

                # Panel 2: Back face
                X = [x_divisions_w1[iw], x_divisions_w1[iw + 1], x_divisions_w2[iw + 1], x_divisions_w2[iw]]
                Y = [h1 / 2, h1 / 2, h2 / 2, h2 / 2]
                Z = [z1, z1, z2, z2]
                nodes0 = np.array([X, Y, Z])
                nodes = np.matmul(R, nodes0) + rA[:, None]
                makePanel(nodes[0], nodes[1], nodes[2], savedNodes, savedPanels)

        for iw in range(nw):
            for ih in range(nh):
                # Panel 3: Left face
                X = [-w1 / 2, -w1 / 2, -w2 / 2, -w2 / 2]
                Y = [y_divisions_h1[ih], y_divisions_h1[ih + 1], y_divisions_h2[ih + 1], y_divisions_h2[ih]]
                Z = [z1, z1, z2, z2]
                nodes0 = np.array([X, Y, Z])
                nodes = np.matmul(R, nodes0) + rA[:, None]
                makePanel(nodes[0], nodes[1], nodes[2], savedNodes, savedPanels)

                # Panel 4: Right face
                X = [w1 / 2, w1 / 2, w2 / 2, w2 / 2]
                Y = [y_divisions_h1[ih], y_divisions_h1[ih + 1], y_divisions_h2[ih + 1], y_divisions_h2[ih]]
                Z = [z1, z1, z2, z2]
                nodes0 = np.array([X, Y, Z])
                nodes = np.matmul(R, nodes0) + rA[:, None]
                makePanel(nodes[0], nodes[1], nodes[2], savedNodes, savedPanels)

    # Mesh the end faces
    wA = new_widths[0]
    hA = new_heights[0]
    wB = new_widths[-1]
    hB = new_heights[-1]
    zA = new_stations[0]
    zB = new_stations[-1]

    # End A face
    nwA = max(int(np.ceil(wA / dw_max)), 1)
    nhA = max(int(np.ceil(hA / dh_max)), 1)
    x_divisions_wA = np.linspace(-wA / 2, wA / 2, nwA + 1)
    y_divisions_hA = np.linspace(-hA / 2, hA / 2, nhA + 1)

    for iw in range(nwA):
        for ih in range(nhA):
            X = [x_divisions_wA[iw], x_divisions_wA[iw + 1], x_divisions_wA[iw + 1], x_divisions_wA[iw]]
            Y = [y_divisions_hA[ih], y_divisions_hA[ih], y_divisions_hA[ih + 1], y_divisions_hA[ih + 1]]
            Z = [zA, zA, zA, zA]
            nodes0 = np.array([X, Y, Z])
            nodes = np.matmul(R, nodes0) + rA[:, None]
            makePanel(nodes[0], nodes[1], nodes[2], savedNodes, savedPanels)

    # End B face
    nwB = max(int(np.ceil(wB / dw_max)), 1)
    nhB = max(int(np.ceil(hB / dh_max)), 1)
    x_divisions_wB = np.linspace(-wB / 2, wB / 2, nwB + 1)
    y_divisions_hB = np.linspace(-hB / 2, hB / 2, nhB + 1)

    for iw in range(nwB):
        for ih in range(nhB):
            X = [x_divisions_wB[iw], x_divisions_wB[iw + 1], x_divisions_wB[iw + 1], x_divisions_wB[iw]]
            Y = [y_divisions_hB[ih], y_divisions_hB[ih], y_divisions_hB[ih + 1], y_divisions_hB[ih + 1]]
            Z = [zB, zB, zB, zB]
            nodes0 = np.array([X, Y, Z])
            nodes = np.matmul(R, nodes0) + rA[:, None]
            makePanel(nodes[0], nodes[1], nodes[2], savedNodes, savedPanels)

    print(f'Of {len(savedPanels) - nSavedPanelsOld} generated panels, {len(savedPanels) - nSavedPanelsOld} were submerged and have been used in the mesh.')

    return savedNodes, savedPanels

def meshRectangularMemberForGDF(stations, widths, heights, rA, rB, dz_max=0, dw_max=0, dh_max=0, endA=True, endB=True):
    '''
    Creates mesh for a rectangular member.

    Parameters
    ----------
    stations: list of locations along member axis at which the cross section will be specified
    widths: list of corresponding widths along member
    heights: list of corresponding heights along member
    rA, rB: member end point coordinates
    dz_max: maximum panel height
    dw_max: maximum panel width
    dh_max: maximum panel height
    endA/endB: flag for whether to mesh each end

    Returns
    -------
    vertices : array
        An array containing the mesh point coordinates, size [3, 4*npanel]
    '''

    if len(stations) != len(widths) or len(stations) != len(heights):
        raise ValueError("The lengths of stations, widths, and heights must be the same.")

    if dz_max == 0:
        dz_max = stations[-1] / 20
    if dw_max == 0:
        dw_max = max(widths) / 8
    if dh_max == 0:
        dh_max = max(heights) / 8

    rAB = np.array(rB) - np.array(rA)
    beta = np.arctan2(rAB[1], rAB[0])
    phi = np.arctan2(np.sqrt(rAB[0]**2 + rAB[1]**2), rAB[2])

    s1 = np.sin(beta)
    c1 = np.cos(beta)
    s2 = np.sin(phi)
    c2 = np.cos(phi)
    s3 = np.sin(0.0)
    c3 = np.cos(0.0)

    R = np.array([[c1 * c2 * c3 - s1 * s3, -c3 * s1 - c1 * c2 * s3, c1 * s2],
                  [c1 * s3 + c2 * c3 * s1, c1 * c3 - c2 * s1 * s3, s1 * s2],
                  [-c3 * s2, s2 * s3, c2]])

    nSavedPanelsOld = 0

    new_stations = [stations[0]]
    new_widths = [widths[0]]
    new_heights = [heights[0]]
    for i in range(len(stations) - 1):
        z1, z2 = stations[i], stations[i + 1]
        w1, w2 = widths[i], widths[i + 1]
        h1, h2 = heights[i], heights[i + 1]

        dz = z2 - z1
        if dz == 0:
            continue

        num_divisions = max(int(np.ceil(dz / dz_max)), 1)
        z_divisions = np.linspace(z1, z2, num_divisions + 1)
        for j in range(1, num_divisions + 1):
            new_stations.append(z_divisions[j])
            new_widths.append(w1 + (w2 - w1) * (z_divisions[j] - z1) / dz)
            new_heights.append(h1 + (h2 - h1) * (z_divisions[j] - z1) / dz)

    x, y, z = [], [], []

    for i in range(len(new_stations) - 1):
        w1, w2 = new_widths[i], new_widths[i + 1]
        h1, h2 = new_heights[i], new_heights[i + 1]
        z1, z2 = new_stations[i], new_stations[i + 1]

        if w1 == 0 or w2 == 0 or h1 == 0 or h2 == 0:
            continue

        if dw_max == 0 or dh_max == 0:
            raise ValueError("dw_max and dh_max must be non-zero.")

        nw = max(int(np.ceil(w1 / dw_max)), int(np.ceil(w2 / dw_max)))
        nh = max(int(np.ceil(h1 / dh_max)), int(np.ceil(h2 / dh_max)))

        x_divisions_w1 = np.linspace(-w1 / 2, w1 / 2, nw + 1)
        x_divisions_w2 = np.linspace(-w2 / 2, w2 / 2, nw + 1)
        y_divisions_h1 = np.linspace(-h1 / 2, h1 / 2, nh + 1)
        y_divisions_h2 = np.linspace(-h2 / 2, h2 / 2, nh + 1)

        for iw in range(nw):
            for ih in range(nh):
                X = [x_divisions_w1[iw], x_divisions_w1[iw + 1], x_divisions_w2[iw + 1], x_divisions_w2[iw]]
                Y = [-h1 / 2, -h1 / 2, -h2 / 2, -h2 / 2]
                Z = [z1, z1, z2, z2]
                nodes0 = np.array([X, Y, Z])
                nodes = np.matmul(R, nodes0) + rA[:, None]
                x.extend(nodes[0])
                y.extend(nodes[1])
                z.extend(nodes[2])

                X = [x_divisions_w1[iw], x_divisions_w1[iw + 1], x_divisions_w2[iw + 1], x_divisions_w2[iw]]
                Y = [h1 / 2, h1 / 2, h2 / 2, h2 / 2]
                Z = [z1, z1, z2, z2]
                nodes0 = np.array([X, Y, Z])
                nodes = np.matmul(R, nodes0) + rA[:, None]
                x.extend(nodes[0])
                y.extend(nodes[1])
                z.extend(nodes[2])

        for iw in range(nw):
            for ih in range(nh):
                X = [-w1 / 2, -w1 / 2, -w2 / 2, -w2 / 2]
                Y = [y_divisions_h1[ih], y_divisions_h1[ih + 1], y_divisions_h2[ih + 1], y_divisions_h2[ih]]
                Z = [z1, z1, z2, z2]
                nodes0 = np.array([X, Y, Z])
                nodes = np.matmul(R, nodes0) + rA[:, None]
                x.extend(nodes[0])
                y.extend(nodes[1])
                z.extend(nodes[2])

                X = [w1 / 2, w1 / 2, w2 / 2, w2 / 2]
                Y = [y_divisions_h1[ih], y_divisions_h1[ih + 1], y_divisions_h2[ih + 1], y_divisions_h2[ih]]
                Z = [z1, z1, z2, z2]
                nodes0 = np.array([X, Y, Z])
                nodes = np.matmul(R, nodes0) + rA[:, None]
                x.extend(nodes[0])
                y.extend(nodes[1])
                z.extend(nodes[2])

    wA = new_widths[0]
    hA = new_heights[0]
    wB = new_widths[-1]
    hB = new_heights[-1]
    zA = new_stations[0]
    zB = new_stations[-1]

    if endA and wA > 0 and hA > 0:
        nwA = max(int(np.ceil(wA / dw_max)), 1)
        nhA = max(int(np.ceil(hA / dh_max)), 1)
        x_divisions_wA = np.linspace(-wA / 2, wA / 2, nwA + 1)
        y_divisions_hA = np.linspace(-hA / 2, hA / 2, nhA + 1)

        for iw in range(nwA):
            for ih in range(nhA):
                X = [x_divisions_wA[iw], x_divisions_wA[iw + 1], x_divisions_wA[iw + 1], x_divisions_wA[iw]]
                Y = [y_divisions_hA[ih], y_divisions_hA[ih], y_divisions_hA[ih + 1], y_divisions_hA[ih + 1]]
                Z = [zA, zA, zA, zA]
                nodes0 = np.array([X, Y, Z])
                nodes = np.matmul(R, nodes0) + rA[:, None]
                x.extend(nodes[0])
                y.extend(nodes[1])
                z.extend(nodes[2])

    if endB and wB > 0 and hB > 0:
        nwB = max(int(np.ceil(wB / dw_max)), 1)
        nhB = max(int(np.ceil(hB / dh_max)), 1)
        x_divisions_wB = np.linspace(-wB / 2, wB / 2, nwB + 1)
        y_divisions_hB = np.linspace(-hB / 2, hB / 2, nhB + 1)

        for iw in range(nwB):
            for ih in range(nhB):
                X = [x_divisions_wB[iw], x_divisions_wB[iw + 1], x_divisions_wB[iw + 1], x_divisions_wB[iw]]
                Y = [y_divisions_hB[ih], y_divisions_hB[ih], y_divisions_hB[ih + 1], y_divisions_hB[ih + 1]]
                Z = [zB, zB, zB, zB]
                nodes0 = np.array([X, Y, Z])
                nodes = np.matmul(R, nodes0) + rA[:, None]
                x.extend(nodes[0])
                y.extend(nodes[1])
                z.extend(nodes[2])

    vertices = np.array([x, y, z])

    vertices2 = np.matmul(R, vertices) + rA[:, None]

    return vertices2.T

def writeMeshToGDF(vertices, filename="platform.gdf", aboveWater=True):

    npan = int(vertices.shape[0]/4)

    f = open(filename, "w")
    f.write('gdf mesh \n')
    f.write('1.0   9.8 \n')
    f.write('0, 0 \n')
    f.write(f'{npan}\n')
    
    if aboveWater:
        for i in range(npan*4):
            f.write(f'{vertices[i,0]:>10.3f} {vertices[i,1]:>10.3f} {vertices[i,2]:>10.3f}\n')
            
    else: # this option avoids making panels above the waterline
    
        for i in range(npan):
            
            panel = vertices[4*i:4*i+4]  # the vertices of this panel
            
            if any(panel[:,2] < -0.001):  # only consider the panel if it's at least partly submerged (some z < 0)
                
                for j in range(4):   # go through each vertex of the panel, but move any above z=0 down to z=0
                
                    if panel[j,2] > 0:  panel[j,2] = 0
                    
                    f.write(f'{panel[j,0]:>10.3f} {panel[j,1]:>10.3f} {panel[j,2]:>10.3f}\n')
    
    f.close()    
    
    

if __name__ == "__main__":

    shape = "rectangular"

    if shape == "circular":
    
        stations = [0, 6, 6,32]
        diameters = [24, 24, 12,  12]
        
        rA = np.array([0, 0,-20])
        rB = np.array([0, 0, 12])
        
        #nodes, panels = meshMember(stations, diameters, rA, rB, dz_max=1, da_max=1)    
        #writeMesh(nodes, panels)
        
        vertices = meshMemberForGDF(stations, diameters, rA, rB, dz_max=5, da_max=5)    
        writeMeshToGDF(vertices)

    elif shape == "rectangular":
        stations= [-120, -12,  -4,  10]
        widths = [24, 24, 24,  24]
        heights = [24, 24, 24,  24]
        
        rA = np.array([0, 0,-10])
        rB = np.array([0, 0, 22])


        vertices = meshRectangularMemberForGDF(stations, widths, heights, rA, rB, dz_max=20, dw_max=20, dh_max=20)
        writeMeshToGDF(vertices)
        savedNodes, savedPanels = meshRectangularMember(stations, widths, heights, rA, rB, dz_max=20, dw_max=20, dh_max=20)
        writeMesh(savedNodes, savedPanels, "Test")



import openmdao.api as om
import os
import yaml

from raft.omdao_raft import RAFT_Group

'''
This example runs RAFT given the exact inputs and options that WEIS provides RAFT in 15_RAFT_Studies
To generate the inputs and options, set the DEBUG_OMDAO flag to true in omdao_raft.py
To debug a specific instance of the opendmdao RAFT run, set the options and input files below using absolute paths
'''

weis_options_file = 'weis_options.yaml'
weis_inputs_file = 'weis_inputs.yaml'

# -----------------------------------
# OMDAO
# -----------------------------------

def test_omdao_raft():
    this_dir = os.path.dirname(__file__)

    # Load options directly generated in WEIS
    with open(os.path.join(this_dir,'test_data',weis_options_file)) as file:
        opt = yaml.load(file, Loader=yaml.FullLoader)

    prob = om.Problem()
    prob.model = RAFT_Group(modeling_options=opt['modeling_options'],
                            analysis_options=opt['analysis_options'],
                            turbine_options=opt['turbine_options'],
                            mooring_options=opt['mooring_options'],
                            member_options=opt['member_options'])
    prob.setup()

    # -------------------------
    # inputs
    # -------------------------
    # Load options directly generated in WEIS
    with open(os.path.join(this_dir,'test_data',weis_inputs_file)) as file:
        inputs = yaml.load(file, Loader=yaml.FullLoader)

    for key, val in inputs.items():
        prob[key] = val


    prob.run_model()

if __name__=="__main__":
    test_omdao_raft()

